package level.blocks;

import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.scene.paint.Color;

/**
 * Builds Blocks from pixels.
 */
public class BlockBuilder {

	/** The prototypes. */
	private static ObservableList<Block> prototypes;

	// Initialize the prototypes
	static {
		prototypes = FXCollections.observableArrayList(new Block(0, 0),
				new TrapBlock(0, 0), new ItemBlock(0, 0), new StandardBlock(0,
						0), new CheckpointBlock(0, 0), new ElevatorBlock(0, 0),
				new GhostBlock(0, 0), new ButtonBlock(0, 0), new PortalBlock(0,
						0));
	}

	/**
	 * Gets the prototypes.
	 * 
	 * @return the prototypes
	 */
	public static ObservableList<Block> getPrototypes() {
		return prototypes;
	}

	/**
	 * Builds a block.
	 * 
	 * @param pX
	 *            the x coordinate of the block
	 * @param pY
	 *            the y coordinate of the block
	 * @param pColor
	 *            the color of the corresponding image pixel
	 * @return the block
	 */
	public static Block build(final int pX, final int pY, final Color pColor) {
		if (pColor.getBrightness() == 0) {
			return new Block(pX, pY);
		}

		int hue = (int) pColor.getHue();

		for (Block prototype : prototypes) {
			if (prototype.isHue(hue)) {
				Block block = prototype.clone();
				block.setX(pX);
				block.setY(pY);
				block.setBrightness(pColor.getBrightness());
				block.setSaturation(pColor.getSaturation());
				block.setTransparency(pColor.getOpacity());
				return block;
			}
		}

		return new Block(pX, pY);
	}

}
