package level.blocks;

import javafx.beans.property.SimpleBooleanProperty;
import javafx.beans.property.SimpleIntegerProperty;
import javafx.collections.FXCollections;
import javafx.scene.Node;
import javafx.scene.control.CheckBox;
import javafx.scene.control.ComboBox;
import javafx.scene.control.Label;
import javafx.scene.control.TitledPane;
import javafx.scene.layout.GridPane;

/**
 * Move up and down periodically or button-triggered.
 */
public class ElevatorBlock extends StandardBlock {
	private static OffsetValueList<Integer> movements = new OffsetValueList<>(
			new OffsetValue<>(0, -4), new OffsetValue<>(3, -3),
			new OffsetValue<>(6, -2), new OffsetValue<>(9, -1),
			new OffsetValue<>(12, 1), new OffsetValue<>(15, 2),
			new OffsetValue<>(18, 3), new OffsetValue<>(21, 4));

	private SimpleBooleanProperty autoMove;

	private SimpleIntegerProperty movement;

	/**
	 * Instantiates a new Elevator block.
	 * 
	 * @param pX
	 *            the x coordinate
	 * @param pY
	 *            the y coordinate
	 */
	public ElevatorBlock(final int pX, final int pY) {
		super(pX, pY);
	}

	@Override
	protected void init() {
		super.init();
		movement = new SimpleIntegerProperty(2);
		movement.addListener(this);
		autoMove = new SimpleBooleanProperty(true);
		autoMove.addListener(this);
	}

	@Override
	protected int getHue() {
		return 180;
	}

	@Override
	protected double getSaturation() {
		int value = 75;
		if (autoMove.get()) {
			value = 50;
		}

		return (value + movements.getByValue(movement.get()).getOffset()) / 100.0;
	}

	@Override
	public void setSaturation(final double pSaturation) {
		int saturation = (int) (100 * pSaturation + 0.5);
		OffsetValue<Integer> i = null;

		if (saturation >= 74) { // ">= 75"
			autoMove.set(false);
			i = movements.getByOffset(saturation - 75);
		} else {
			autoMove.set(true); // ">= 50" + fallback
			i = movements.getByOffset(saturation - 50);
		}

		if (i == null) {
			autoMove.set(true);
			movement.set(2);
		} else {
			movement.set(i.getValue());
		}
		draw();
	}

	@Override
	protected boolean needsId() {
		return true;
	}

	@Override
	protected ElevatorBlock subClone() {
		return new ElevatorBlock(getX(), getY());
	}

	@Override
	public String toString() {
		return "Elevator Block [" + movement.get() + "] at (" + getX() + "|"
				+ getY() + ")";
	}

	@Override
	public String getTypeString() {
		return "Elevator";
	}

	@Override
	protected Node getSaturationMenu() {
		TitledPane saturationMenu = new TitledPane();
		saturationMenu.setCollapsible(false);
		saturationMenu.setText("Elevator properties");

		GridPane controls = new GridPane();
		controls.getStyleClass().add("grid-pane");

		final CheckBox autoMoveCheckBox = new CheckBox("Move automatically");
		autoMoveCheckBox.setSelected(autoMove.get());
		autoMove.bind(autoMoveCheckBox.selectedProperty());
		controls.add(autoMoveCheckBox, 0, 1);
		GridPane.setColumnSpan(autoMoveCheckBox, 2);

		controls.add(new Label("Movement:"), 0, 2);

		ComboBox<Integer> movementCombo = new ComboBox<>(
				FXCollections.observableArrayList(-4, -3, -2, -1, 1, 2, 3, 4));
		movementCombo.getSelectionModel().select(movement.getValue());
		movement.bind(movementCombo.valueProperty());
		controls.add(movementCombo, 1, 2);

		saturationMenu.setContent(controls);

		return saturationMenu;
	}
}
