package level.blocks;

/**
 * A mapping of saturation offsets and corresponding values.
 * 
 * @param <T>
 *            the value type
 */
public class OffsetValue<T extends Comparable<T>> {

	/** The offset (on the saturation value). */
	private final int offset;

	/**
	 * Gets the offset (on the saturation value).
	 * 
	 * @return the offset (on the saturation value)
	 */
	public int getOffset() {
		return offset;
	}

	/** The value. */
	private final T value;

	/**
	 * Gets the value.
	 * 
	 * @return the value
	 */
	public T getValue() {
		return value;
	}

	/**
	 * Instantiates a new mapping.
	 * 
	 * @param pOffset
	 *            the offset on the saturation value
	 * @param pValue
	 *            the value
	 */
	public OffsetValue(final int pOffset, final T pValue) {
		offset = pOffset;
		value = pValue;
	}

	/**
	 * Checks if the specified offset should be interpreted as this mapping's
	 * offset.
	 * 
	 * @param pOffset
	 *            the offset
	 * @return true, if the offset is in +-1 of this mapping's offset
	 */
	public boolean isOffset(final int pOffset) {
		return (Math.abs(offset - pOffset) <= 1);
	}

	/**
	 * Checks if the specified value should be interpreted as this mapping's
	 * value.
	 * 
	 * @param pValue
	 *            the value
	 * @return true, if the two values are equal according to
	 *         {@link T#compareTo(Comparable)}
	 */
	public boolean isValue(final T pValue) {
		return value.compareTo(pValue) == 0;
	}
}
